#pragma once

#include <vector>
#include <math.h>

// on error std::runtime_error is thrown by value
// implementation of median of medians algorithm, described here :
// https://en.wikipedia.org/wiki/Median_of_medians

template <class TType>
class CMedian
{
typedef std::vector<TType> TTypeVctr;

public:
	CMedian(double fCMPPrecision);
	virtual ~CMedian();

	void SetDataset(const TTypeVctr &v);
	void AddDataset(const TTypeVctr &v);
	void AddDatasetPoint(const TType &v);

	TType CalcMedian(bool bDumpProfileData);

protected:
	size_t SelectTop(size_t left, size_t right, size_t n);

	size_t Partition(size_t left, size_t right, size_t pivotIndex, size_t n);
	size_t Partition5(size_t left, size_t right);
	size_t Select(size_t left, size_t right, size_t n);
	size_t FindPivotIndex(size_t left, size_t right);

// members
protected:
	double m_fCMPPrecision;		// comparison precision
	TTypeVctr m_DataSet;		// Current dataset, may/will differ from initial set

	size_t m_nTopSelectIterations;
	size_t m_nSelectCallsCount;
	size_t m_nSelectRecursionCount;
	size_t m_nSelectRecursionCountMax;
};


#include "Median.hpp"
