//#include "stdafx.h"

#include <cstring>
#include <stdio.h>
#include <iostream>
#include <vector>
#include <time.h>
#include <algorithm>
#include <assert.h>
#include <stdint.h>

#include "Median.h"

using namespace std;

typedef long long int INT64;
typedef unsigned long long int UINT64;

// These control all test parameters
//////////////////////////////////////////////////////////////////////////
#if 1
// with float
//////////////////////////////////////////////////////////////////////////
#define TESTS_COUNT				70			/* Repeat count of each type of test */
#define DATASET_ELEMENTS		100000		/* Begin set count, dataset for each test is TESTS_COUNT * DATASET_ELEMENTS */
#define COMPARE_PRECISION		0.0001		/* If distance between 2 numbers is less than this, they are considered equal */
#define VERB_OUTPUT				false		/* true/false - will dump more statistics for algorithm */
typedef double TDataType;					/* works with signed types only [float, double] */
//////////////////////////////////////////////////////////////////////////
#endif

#if 0
// with integer
//////////////////////////////////////////////////////////////////////////
#define TESTS_COUNT				70			/* Repeat count of each type of test */
#define DATASET_ELEMENTS		100000		/* Begin set count, dataset for each test is TESTS_COUNT * DATASET_ELEMENTS */
#define COMPARE_PRECISION		1			/* If distance between 2 numbers is less than this, they are considered equal */
#define VERB_OUTPUT				false		/* true/false - will dump more statistics for algorithm */
typedef INT64 TDataType;						/* works with signed types only [char, short, int, INT64] char & short have some overflow problems, depending of test data, no time to track */
//////////////////////////////////////////////////////////////////////////
#endif

typedef vector<TDataType> TDataVector; 
typedef TDataVector::iterator TDataIt;

enum EDataTestType { DTT_RANDOM = 0, DTT_SORT_ASC, DTT_SORT_DESC, DTT_SAME_VALUES };

#if defined(_WIN32) || defined(_WIN64)

extern "C" unsigned __int64 __stdcall GetTickCount64();

	UINT64 GetTickCount()
	{
		return ::GetTickCount64();
	}

#else	// Linux

	UINT64 GetTickCount()
	{
		struct timespec ts;
		memset(&ts, 0x00, sizeof(ts));

		if(clock_gettime(CLOCK_MONOTONIC, &ts) != 0) 
		{
			assert(false);
		}

		UINT64 nRes = ts.tv_sec * 1000;	// in ms
		nRes += ts.tv_nsec / 1000000;	// extract ms

		return nRes;
	}
#endif

// sort & get median
TDataType GetMedianClassic_stdsort(TDataVector &DataSet)
{
	if(!DataSet.size())
		throw runtime_error("GetMedianClassic invalid dataset!");

	sort(DataSet.begin(), DataSet.end());
	if(DataSet.size() & 1)
		return DataSet[DataSet.size() / 2];		// element count odd, so get middle array value
	else
		return (DataSet[DataSet.size() / 2] + DataSet[DataSet.size() / 2 - 1]) / 2;		// element count even, so get mean of middle 2 array elements
}

int qsrtfn_asc(const void *p1, const void *p2)
{
	TDataType dt = (*((TDataType *) p1) - *((TDataType *) p2));
	if(dt == 0)
		return 0;

	return dt > 0 ? 1 : -1;
}

int qsrtfn_desc(const void *p1, const void *p2)
{
	TDataType dt = (*((TDataType *) p2) - *((TDataType *) p1));
	if(dt == 0)
		return 0;

	return dt > 0 ? 1 : -1;
}

// sort & get median
TDataType GetMedianClassic_qsort(TDataVector &DataSet)
{
	if(!DataSet.size())
		throw runtime_error("GetMedianClassic invalid dataset!");

	qsort(&(*DataSet.begin()), DataSet.size(), sizeof(TDataType), qsrtfn_asc);

	if(DataSet.size() & 1)
		return DataSet[DataSet.size() / 2];		// element count odd, so get middle array value
	else
		return (DataSet[DataSet.size() / 2] + DataSet[DataSet.size() / 2 - 1]) / 2;		// element count even, so get mean of middle 2 array elements
}

void GenerateTestData(EDataTestType nDataSetType, size_t nDatasetPoints, TDataVector &DataSet)
{
	if(!nDatasetPoints)
		throw runtime_error("GenerateTestData invalid dataset points count!");

	DataSet.resize(nDatasetPoints);

	if(nDataSetType == DTT_SAME_VALUES)
	{
		TDataType DataVal = (TDataType) (rand() * COMPARE_PRECISION);
		DataSet.assign(nDatasetPoints, DataVal);
		return;
	}

	// rand() & 0xFFFF - limit values, qsort & maybe other problems with overflow
	generate(DataSet.begin(), DataSet.end(), [](){ return ((TDataType) (rand() & 0xFFFF)) * (rand() % 2 ? COMPARE_PRECISION : -COMPARE_PRECISION); });
	switch(nDataSetType)
	{
		case DTT_RANDOM :		break;
		case DTT_SORT_ASC :		qsort(&(*DataSet.begin()), DataSet.size(), sizeof(TDataType), qsrtfn_asc);		break;
		case DTT_SORT_DESC :	qsort(&(*DataSet.begin()), DataSet.size(), sizeof(TDataType), qsrtfn_desc);		break;
		default:
			{
				assert(false);	// add appropriate type here
			}
	}
}

void TestMedian(TDataVector &TestDataset)
{
	TDataVector ds;
	UINT64 nStartTm, nDurationTm;
	printf("%lu, ", TestDataset.size());

	// Calc by median of medians approach
	CMedian<TDataType> mc(COMPARE_PRECISION);	// work with 
	mc.SetDataset(TestDataset);
	nStartTm = GetTickCount();	
	TDataType Median1 = mc.CalcMedian(VERB_OUTPUT);		// Will calculate it
	nDurationTm = GetTickCount() - nStartTm;
	printf("%llu, ", nDurationTm);

	// Calc by std::sort & get middle element
	ds = TestDataset;
	nStartTm = GetTickCount();	
	TDataType Median2 = GetMedianClassic_stdsort(ds);
	nDurationTm = GetTickCount() - nStartTm;
	printf("%llu, ", nDurationTm);

	// Calc by qsort & get middle element
	ds = TestDataset;
	nStartTm = GetTickCount();	
	TDataType Median3 = GetMedianClassic_qsort(ds);
	nDurationTm = GetTickCount() - nStartTm;
	printf("%llu\n", nDurationTm);

	// Test if equal
	if(std::abs(Median1 - Median2) >= COMPARE_PRECISION)
	{
//printf("Median1 = %d, Median2 = %d, Median3 = %d\n", Median1, Median2, Median3);		
		assert(false);
		throw runtime_error("Algorithm comparison 1 (m_of_m vs qsort) failed!");
	}

	// qsort does not work for integer types less than int
//	if(sizeof(TDataType) >= 4)
	{
		if(std::abs(Median1 - Median3) >= COMPARE_PRECISION)
		{
			assert(false);
			throw runtime_error("Algorithm comparison 2 (m_of_m vs qsort) failed!");
		}
	}

	// too slow for tests, but works
	if(false)
	{
		// Test add median point, add another set, point by point
		for(size_t i = 0; i < 50; i++)
		{
			TDataType NewVal = (TDataType) (rand() / 10.0);

			mc.AddDatasetPoint(NewVal);
			Median1 = mc.CalcMedian(false);

			ds = TestDataset;
			ds.push_back(NewVal);
			Median2 = GetMedianClassic_stdsort(ds);

			if(std::abs(Median2 - Median1) >= COMPARE_PRECISION)
				throw runtime_error("Algorithm comparison failed!");
		}
	}
}

void TestMedianWithDataset(EDataTestType nDataSetType, const char *szDataSetType)
{
	printf("\n\n Test with dataset : %s \n====================================================================================\n", szDataSetType);
	printf("Test #, DataSet count, m of m [ms], classic by std::sort [ms], classic by qsort [ms]\n");

	TDataVector TestDataset;

	for(int i = 1; i <= TESTS_COUNT; i++)
	{
		printf("%d, ", i);

		// randomize set count, so we can test even/odd number of elements & left overs calculations in algorithm
		size_t nDatasetPoints = (DATASET_ELEMENTS * i) + (rand() % 100);
		GenerateTestData(nDataSetType, nDatasetPoints, TestDataset);
		TestMedian(TestDataset);
	}
}

int main(int argc, char* argv[])
{
	srand((unsigned int) time(NULL));
	int nResult = EXIT_FAILURE;

	cout << "TDataType size : " << sizeof(TDataType) << endl;
	
	try
	{
// For manual tests
//////////////////////////////////////////////////////////////////////////
/*
	TDataVector v, ds;
	size_t nDatasetPoints = 10;
	GenerateTestData(DTT_RANDOM, nDatasetPoints, v);
	
ds = v;	
TDataType Median1 = GetMedianClassic_stdsort(ds);
ds = v;	
TDataType Median2 = GetMedianClassic_qsort(ds);

	CMedian<TDataType> mc(COMPARE_PRECISION);
	mc.SetDataset(v);
	TDataType Median3 = mc.CalcMedian(true);		// Will calculate it
	printf("done\n");

	return 0;
*/
//////////////////////////////////////////////////////////////////////////

		TestMedianWithDataset(DTT_RANDOM, "Random values");
		TestMedianWithDataset(DTT_SORT_ASC, "Random values sorted (asc)");
		TestMedianWithDataset(DTT_SORT_DESC, "Random values sorted (desc)");
		TestMedianWithDataset(DTT_SAME_VALUES, "Single random value");

		printf("\n\n====================================================================================\n");
		printf("\n\nTEST SUCCESS!!!\n");
		printf("\n\n====================================================================================\n");

		nResult = EXIT_SUCCESS;
	}
	catch(const runtime_error &ex)
	{
		printf("TEST FAIL : %s\n", ex.what());
	}
	catch(const runtime_error *pex)
	{
		printf("TEST FAIL : %s\n", pex->what());
		delete pex;
	}
	catch(...)
	{
		printf("TEST FAIL : Unhandled exception!\n");
	}

	return nResult;
}

